#!/usr/bin/perl
$| = 1;
use IO::Socket;
use strict;
use MiniServer;
use WebFTPPages;
use Net::FTP;

 #############
# Global vars #
###############
my %conf;
load_conf(\%conf);
my($FTPhost,$FTPhostname);
my $Version = '1.5.2';

@{$conf{h}} = split /\s*,\s*/, $conf{ftphost};
@{$conf{hn}} = split /\s*,\s*/, $conf{ftphostname};
for(my $i=0; $i<=$#{$conf{h}}; $i++) {
    $conf{hn}->[$i] ||= $conf{h}->[$i];
    $conf{whathostname}{$conf{h}->[$i]} = $conf{hn}->[$i];
}

unless($#{$conf{h}}) {
    $FTPhost = $conf{h}->[0];
    $FTPhostname = $conf{hn}->[0];
}

# Don't alter these #
#  unless you know  #
# what you're doing #

  
$FTPhostname ||= $FTPhost;
my $BufferSize = 5 * 1024;
my $CRLF = "\015\012";
my $RemoteHost = $ENV{'REMOTE_ADDR'};  # These may have to change if
my $Scriptname = $ENV{'SCRIPT_NAME'};  # you're not using Apache.
my $Servername = $ENV{'HTTP_HOST'};    #
my ($ServerPort,$FTP);
$conf{SCRIPT_NAME} = $Scriptname;
$conf{HTTP} = (($ENV{HTTPS} =~ /on/i )? 'https://' : 'http://');
 ##############
# Main program #
################

my($CONTENT,$EDITOR,$HTTPD);

if($ENV{PATH_INFO}) {
    &HandleSession;
} else {
    &UserLogin;
}
 ###############
# Login handler #
#################
sub UserLogin {
    my($host, $username, $password);
    
    while (<STDIN>) {
	if(/(?:host=(\S*)&)?username=(\S*)&password=(\S*)/) {
	    $FTPhost = $FTPhostname = $1 if $1;
	    $FTPhostname = $conf{whathostname}{$FTPhost};
	    $FTPhost ||= 'localhost';
	    $username = $2;
	    $password = $3;
	}
	if($_ eq 'CRLF') { last; }
    }
    #ANT check this out, this gets run on startup?
    $username =~ s/\%([A-Fa-f0-9]{2})/chr(hex($1))/eg if($username);
    $password =~ s/\%([A-Fa-f0-9]{2})/chr(hex($1))/eg if($password);
    
    $conf{username} = $username;
    $conf{password} = $password;
    
    $CONTENT = WebFTPPages->new($Version,
				$FTPhost,
				$FTPhostname,
				$Servername,
				$Scriptname,
				\%conf);
    
    $HTTPD = MiniServer->new($RemoteHost,
			     $CONTENT,
			     $Servername,
			     \&FileMaster,
			     \%conf);
    
    $conf{HTTPD} = $HTTPD;
    $conf{CONTENT} = $CONTENT;
    
    if($username && $password) { 
	if(&TryFTPLogin([$username, $password, 1])) {
	    #ANT: this is a major kludge... STDOUT not closing makes some
	    # browsers hang and not redirect. However, closing STDOUT causes
	    # apache to kill the cgi. SO! I must fork and close the parent,
	    # then close STDOUT thus saving the day... till I find a better
	    # solution
	    $HTTPD->WANT_COOKIE(1);
#          &TryFTPLogin([$username, $password]);
	    close STDOUT;
	    &Serve;
	}
    } else { 
	$HTTPD->sendResponse($CONTENT->loginScreen()); 
	exit(0); 
    }
}

sub TryFTPLogin {
  my ($user, $pass, $test) = @{$_[0]}; #passed this way to hide password from logs, sigh

  unless($FTP = Net::FTP->new($FTPhost, Timeout => $conf{timeout}+50)) {
      $HTTPD->sendResponse($CONTENT->errorPage("Server said: ".$FTP->message));
      die "Server said: ".$FTP->message;
  }
  unless( $FTP->login($user, $pass) ) {
      if($FTP->message =~ /CRYPTOCard challenge (\d+)/i) {
	  $conf{challenge} = $1;
      } else {
	  $FTP->quit;
	  $HTTPD->sendResponse($CONTENT->failLogin());
	  exit(0);
      }
  }
#  return 1 if $test;
  if(fork()) {
      $FTP->close();
      exit(0);
  }
     
#  $FTP->cwd($conf{initialdir}) if $conf{initialdir};
  # All systems are go. Let's start the webserver.
  $conf{FTP} = $FTP;
  $HTTPD->startServer;
  $HTTPD->sendRedirect("$conf{HTTP}$Servername$Scriptname/START?$$+$conf{challenge}");
}

 #############
# Server core #
###############
sub Serve {
  $SIG{ALRM} = \&Timeout;
  alarm($conf{timeout});

  while(1) {
    unless($HTTPD->awaitConnection()) { next; }
    my $queryHashRef = $HTTPD->handleQueryHeaders();
    next unless($queryHashRef);
    if($queryHashRef->{REQUEST_METHOD} eq 'GET') { &GetHandler($queryHashRef->{PATH}); }
    elsif($queryHashRef->{REQUEST_METHOD} eq 'POST') { &PostHandler($queryHashRef); }
    
    $HTTPD->closeConnection unless $conf{noclose};
    $conf{noclose} = 0;
  }
}

sub GetHandler {
    $_ = shift;
    s|^/(?:\Q$conf{SCRIPT_NAME}\E/)*||;
    if(($_ eq '') || ($_ eq 'LIST')) {
        $CONTENT->updateDir($FTP->pwd());
        $CONTENT->getDirectory($FTP->dir());
	$HTTPD->genCookie();
        $HTTPD->sendResponse($CONTENT->frameset());
    }
    elsif($_ eq 'PUT') { $HTTPD->sendResponse($CONTENT->uploadScreen()); }
    elsif($_ eq 'MKDIR') { $HTTPD->sendResponse($CONTENT->mkdirScreen()); }
    elsif($_ eq 'CDUP') {
        $FTP->cdup;
	unshift @{$conf{messages}}, [time(), $FTP->message];
        alarm($conf{timeout});
        $HTTPD->sendRedirect("$conf{HTTP}$Servername$conf{SCRIPT_NAME}$conf{COOKIE}/");
    }
    elsif($_ eq 'QUIT') {
        $FTP->quit;
        $HTTPD->sendRedirect("$conf{HTTP}$Servername$Scriptname");
        cleanup();
    } elsif(/CDTO\?(.*)/) {
        $FTP->cwd($CONTENT->unencode($1));
	unshift @{$conf{messages}}, [time(), $FTP->message];
        alarm($conf{timeout});
        $CONTENT->updateDir($FTP->pwd());
        $CONTENT->getDirectory($FTP->dir());
        $HTTPD->sendResponse($CONTENT->MAIN());
    } elsif(/PERM\?(.*)/) {
        $HTTPD->sendResponse($CONTENT->permsScreen($1));
    } elsif(/DNLD\?(.*)/) {
        my(%foo);
        $foo{Files} = [$1];
        $foo{action} = ['download'];
	my $pwd = $FTP->pwd();

	if($conf{challenge}) {
	    download(\%foo);
	} elsif (!fork()) {
	    close $FTP; # Much thanks to Grahamm Barr for this fix.
	    unless($FTP = Net::FTP->new($FTPhost, 
					Timeout => $conf{timeout}+50)) {
		$HTTPD->sendResponse($CONTENT->errorPage("Server said: ".$FTP->message));
		die "Server said: ".$FTP->message;
	    }

	    $FTP->login($conf{username}, $conf{password});
	    unshift @{$conf{messages}}, [time(), $FTP->message];
	    $FTP->cwd($pwd);
	    unshift @{$conf{messages}}, [time(), $FTP->message];
	    download(\%foo);

	    $HTTPD->closeConnection();
	    exit(0);
	}
    } elsif(/EDIT\?(.*)/) {
        my(%foo);
	my $str = $1;
	$str =~ /file=([^&]+)/;
        $foo{Files} = [$1];
        $foo{action} = ['edit'];
	$str =~ /editor=([^&]+)/;
	my $editor = $1;
	require "$editor.pm";
	$EDITOR = eval "new $editor('conf' => \\\%conf)";
        download(\%foo);
    } elsif(/MAIN/) {
	$CONTENT->updateDir($FTP->pwd());
	$CONTENT->getDirectory($FTP->dir());
	$HTTPD->sendResponse($CONTENT->MAIN());
    } elsif($_) {
	my $str;
	s/\?(.*)$//;
        eval {
            $str = eval "\$CONTENT->$_(\$1)";
	    die if $@;
        };
	if($@) {
	    eval {
		$str = eval "\$EDITOR->$_(\$1)";
	    };
	    die if $@;
	}
#	unshift @{$conf{messages}}, [time(), $FTP->message];
        $HTTPD->sendResponse($str) if($str && !$@);
    }
}
BEGIN { # a horrendous kludge
    my $FILE;

  sub FileMaster {
    my $directive = shift;
    my $full_path = $_[1]; # Ewwwww ugly. # ant

    if($directive eq 'NEW') {
      my $filename = shift;
      $filename =~ s/^.*[\\\/\:]// unless $full_path;
      $FTP->binary;
      $FILE = $FTP->stor($filename);
      unshift @{$conf{messages}}, [time(), $FTP->message()];
    }
    elsif($directive eq 'PUT') {
      my $buff = shift;
      if($FILE && length($buff)) { $FILE->write($buff, length($buff)); }
    }
    elsif($directive eq 'CLOSE') {
	if($FILE) {
	    $FILE->close;
	    $FILE = undef;
	    alarm($conf{timeout});
	} else {
	    $conf{save_error} = 1;
	    unshift @{$conf{messages}}, [time(), "Could not write file: ". $FTP->message()];
	}
    }
  }
}

sub PostHandler {
  my $QueryHashRef = shift;

  my @keys = keys %$QueryHashRef;

  my @Files = @{$QueryHashRef->{'Files'}}
                                       if(exists($QueryHashRef->{'Files'}));
  my @Directories;
  @Directories = @{$QueryHashRef->{'Directories'}}
                                 if(exists($QueryHashRef->{'Directories'}));
  push @Directories, split "\n", $QueryHashRef->{'DirectoriesList'}->[0]
                                 if($QueryHashRef->{'DirectoriesList'}->[0]);
  my($Action) = @{$QueryHashRef->{'action'}}
                                      if(exists($QueryHashRef->{'action'}));

  if($Action eq 'download') {
      return &download($QueryHashRef);
  }
  elsif($Action eq 'site') {
      $FTP->quot('SITE',$Directories[0]);      
  }
  elsif($Action eq 'edit') {
      $conf{EDITDATA} = $QueryHashRef->{data}[0];
      doEditSave();
      $HTTPD->sendResponse($EDITOR->SYS);
      return;
  }
  elsif($Action eq 'move') {
      $HTTPD->sendResponse($CONTENT->moveScreen(@Directories, @Files));
      return;
  }
  elsif($Action eq 'Move') {
      while($_ = shift @Files) {
          $FTP->rename($_, shift @Directories);
          alarm($conf{timeout});
      }
  }
  elsif($Action eq 'permissions') {
      $HTTPD->sendResponse($CONTENT->permsScreen(@Directories, @Files));
      return 0;
  }
  elsif($Action eq 'Change permissions') {
      my $i = 0;
      while($_ = shift @Files) {
          my $newperms = shift(@Directories);	  
	  if($newperms) {
	      if($newperms =~ /[0-7]{3,4}/) {
		  $FTP->quot('SITE',"CHMOD $newperms $_");
	      }
	  } else {
	      $newperms = 0;
	      if($QueryHashRef->{"perm_$i"}) {
		  for(@{$QueryHashRef->{"perm_$i"}}) {
		      $newperms |= $_;
		  }
		  $FTP->quot('SITE',sprintf("CHMOD %03o $_", $newperms));
	      }
	  }
	  $i++;
	  alarm($conf{timeout});
      }
  } elsif($Action eq 'chdir') {
      $FTP->cwd(shift @Directories);
      alarm($conf{timeout});
  } elsif($Action eq 'Make directory') {
      $FTP->mkdir(shift @Directories, 1);
      alarm($conf{timeout});
  }
  elsif($Action eq 'delete') {
      while($_ = shift @Files) { $FTP->delete($CONTENT->unencode($_)); alarm($conf{timeout}); }
      while($_ = shift @Directories) { $FTP->rmdir($CONTENT->unencode($_)); alarm($conf{timeout}); }
  } elsif($Action eq 'Upload') {
  } elsif($Action eq 'crypto') {
      my $m;
      $FTP->quot("acct", $QueryHashRef->{response}[0]);
      $m = $FTP->message;
      if($m =~ /access denied/i) {
	  $HTTPD->sendResponse($CONTENT->errorPage("Server said: $m"));	  
      } else {
	  $HTTPD->sendResponse($CONTENT->frameset());
      }
      return;
  } elsif($Action) {
      my $str;
      s/\?(.*)$//;
      eval {
	  $str = eval "\$EDITOR->$Action(\$QueryHashRef,\$1)";
      };
      if($conf{state}) {
	  if($conf{state} eq 'save') {
	      doEditSave();
	      $conf{state} = '';
	  }
      }
      unless($@) {
	  $HTTPD->sendResponse(eval "\$EDITOR->$str()");
	  return 0;
      }
  }
  unshift @{$conf{messages}}, [time(), $FTP->message];
  # nothing much happens for the 'Upload' action
  $CONTENT->updateDir($FTP->pwd());
  $CONTENT->getDirectory($FTP->dir());
  $HTTPD->sendResponse($CONTENT->MAIN());
}

sub doEditSave {
    &FileMaster('NEW', $EDITOR->get_filename(),1);
    &FileMaster('PUT', $EDITOR->get_data());
    &FileMaster('CLOSE');
}

sub download {
    my $QueryHashRef = shift;
    my @Files = @{$QueryHashRef->{'Files'}}
      if(exists($QueryHashRef->{'Files'}));
    my @Directories = @{$QueryHashRef->{'Directories'}}
      if(exists($QueryHashRef->{'Directories'}));
    my($Action) = @{$QueryHashRef->{'action'}}
      if(exists($QueryHashRef->{'action'}));
    my $FTPbuffer;

    # Sigh - we have to cheat and absorb some server functions here.
    my $CLIENT = $HTTPD->Client;
    alarm($conf{timeout});
    $FTP->binary;
    foreach my $file (@Files) { 
	$file = $CONTENT->unencode($file);
	my $size = $FTP->size($file);
	unshift @{$conf{messages}}, [time(), $FTP->message];
	if($Action eq 'download') {
	    print $CLIENT qq`Content-Type: application/octet-stream\n`.
		qq`Connection: close\n` .
			qq`Content-Disposition: attachment;filename="$file"\n`.
			    qq`Content-Length: $size\n\n`;
	} else {
	   print $CLIENT qq`Content-Type: text/html\n\n
<HTML><BODY><SCRIPT LANGUAGE=JAVASCRIPT>
alert("File $file is too big to edit/view. File may be no larger than $conf{maxeditsize} bytes");
self.close();
</SCRIPT></BODY></HTML>` if $size > $conf{maxeditsize};
	   return if $size > $conf{maxeditsize};
	   $conf{EDITDATA} = '';
	   if($file =~ m|^/|) {
	       $conf{CURRENTEDIT} = $file;
	   } else {
	       $conf{CURRENTEDIT} = $CONTENT->getDir."/$file";
	   }
	}
	my $FILE = $FTP->retr($file);
	unshift @{$conf{messages}}, [time(), $FTP->message];
	if(defined($FILE)) {
	    while(my $bytecount = $FILE->read($FTPbuffer, $BufferSize)) {
		alarm($conf{timeout});
		if($Action eq 'download') {
		    print $CLIENT $FTPbuffer;
		} else {
#		    $conf{EDITCWD} .= $FTP->pwd();
		    $conf{EDITDATA} .= $FTPbuffer;
		}
	    }
	    $FILE->close;
	} else {
	    $EDITOR->new_file();
	}
	alarm($conf{timeout});
	if($Action ne 'download') {
	    $HTTPD->sendResponse($EDITOR->EDITMAIN());
	    return;
	}
    }
    return;
}

sub Timeout {
  $FTP->quit;
  warn "Remote user timed out";
  cleanup();
}

sub load_conf {
    my($key,$value);
    open(DATA, 'web-ftp.conf') or warn 'No web-ftp.conf file';
    while(<DATA>) {
        next if ord($_) == 35;
        next unless /\s*(\w+)\s*=\s*(.*)/;
        $key = lc($1);
        if($value = $2) {
            $value =~ s/^\s*[yY][eE][sS]\s*$/1/;
            $value =~ s/^\s*[nN][oO]\s*$/0/;
        }
	if($key eq 'editors') {
	    my @v = split ',', $value;
	    for(@v) { s/^\s+//; s/\s+$//; }
	    push @{$_[0]->{$key}}, \@v;
	} else {
	    $_[0]->{$key} = $value;
	}
    }

    if($_[0]->{hidetag}) {
        $_[0]->{hidedownload} = 1;
        $_[0]->{hidedelete} = 1;
        $_[0]->{hidemove} = 1;
        $_[0]->{hidepermbutton} = 1;
    }
    $_[0]->{hidepermbutton} = 1 if $_[0]->{hidepermissions};

    $_[0]->{ftphost} = 'localhost' unless defined $_[0]->{ftphost};
    $_[0]->{ftphostname} = 'Localhost' unless defined $_[0]->{ftphostname};
    $_[0]->{timeout} ||= 300;
    $_[0]->{backgroundcolor} = '#FFFFFF' unless defined $_[0]->{backgroundcolor};
    $_[0]->{buttoncolor} = '#CCAAAA' unless defined $_[0]->{buttoncolor};
    $_[0]->{errorcolor} = '#FF0000' unless defined $_[0]->{errorcolor};
    $_[0]->{linkcolor} = '#0000FF' unless defined $_[0]->{linkcolor};
    $_[0]->{logincolor} = '#77EECC' unless defined $_[0]->{logincolor};
    $_[0]->{textcolor} = '#000000' unless defined $_[0]->{textcolor};
    $_[0]->{visitedlinkcolor} = '#990099' unless defined $_[0]->{visitedlinkcolor};
    $_[0]->{oddcolcolor} = '#F0F0F0' unless defined $_[0]->{oddcolcolor};
    $_[0]->{evencolcolor} = '#FFFFFF' unless defined $_[0]->{evencolcolor};
    $_[0]->{hidedelete} = '0' unless defined $_[0]->{hidedelete};
    $_[0]->{hidedownload} = '0' unless defined $_[0]->{hidedownload};
    $_[0]->{hidelastmod} = '0' unless defined $_[0]->{hidelastmod};
    $_[0]->{hidemkdir} = '0' unless defined $_[0]->{hidemkdir};
    $_[0]->{hidemove} = '0' unless defined $_[0]->{hidemove};
    $_[0]->{hideowngrp} = '0' unless defined $_[0]->{hideowngrp};
    $_[0]->{hidepermbutton} = '0' unless defined $_[0]->{hidepermbutton};
    $_[0]->{hidepermissions} = '0' unless defined $_[0]->{hidepermissions};
    $_[0]->{hidetag} = '0' unless defined $_[0]->{hidetag};
    $_[0]->{hidesize} = '0' unless defined $_[0]->{hidesize};
    $_[0]->{hidemessages} = '0' unless defined $_[0]->{hidemessages};
    $_[0]->{hidesite} = '1' unless defined $_[0]->{hidesite};
    $_[0]->{hideupload} = '0' unless defined $_[0]->{hideupload};    
    $_[0]->{sockdir} = '/tmp/Web-FTP' unless defined $_[0]->{sockdir};    
    $_[0]->{mimefile} = '/etc/mime.types' unless defined $_[0]->{mimefile};
    $_[0]->{maxeditsize} = '100000' unless defined $_[0]->{maxeditsize};
    $_[0]->{fileframesize} = '*' unless defined $_[0]->{fileframesize};
    $_[0]->{dirsframesize} = '*' unless defined $_[0]->{dirsframesize};
    $_[0]->{infoframesize} = '40' unless defined $_[0]->{infoframesize};

#    $_[0]->{} = '' unless defined $_[0]->{};   

    
}

sub cleanup() {
    $HTTPD->closeServer();
}

sub HandleSession {
    my($dir,$s,$peer);

    $ENV{PATH} = "$ENV{PATH_INFO}";
    $ENV{PATH} .= "?$ENV{QUERY_STRING}" if $ENV{QUERY_STRING};
   
#    if($ENV{PATH_INFO} eq '/ACCT') {
#	my $tmp;
#	($peer,$tmp) = split /\+/, $ENV{QUERY_STRING};
#	$ENV{PATH} = "$ENV{PATH_INFO}";
#	$ENV{PATH} .= "?$tmp";
    if($ENV{PATH_INFO} eq '/START') {
#        $peer = $ENV{QUERY_STRING};
	my $tmp;
	($peer,$tmp) = split /\+/, $ENV{QUERY_STRING};
        $ENV{PATH} = '';
	if($tmp) {
	    $conf{challenge} = $tmp;
	    $ENV{PATH} = 'ACCT' if $conf{challenge};
	}
    } else {
	if($conf{urlsessions}) {
	    $ENV{PATH_INFO} =~ s/^\/([^\/]+)//;
	    $ENV{HTTP_COOKIE} = "WEBFTPTAG=$1";
	    $ENV{PATH} = "$ENV{PATH_INFO}";
	    $ENV{PATH} .= "?$ENV{QUERY_STRING}" if $ENV{QUERY_STRING};
	}
        $ENV{HTTP_COOKIE} =~ /WEBFTPTAG=[^+]+\+(\d+)/;
        $peer = $1;
    }

    $s = new IO::Socket::UNIX('Peer' => "$conf{socketdir}/$peer");
    unless($s) {
        print qq`Content-type: text/html

<HTML><HEAD>
<SCRIPT LANGUAGE=JAVASCRIPT>
function goToLogin() {
  setTimeout("open('$Scriptname','_top')",5000);
}
</SCRIPT>
</HEAD>
<BODY BGCOLOR="$conf{backgroundcolor}" TEXT="$conf{textcolor}" VLINK="$conf{visitedlinkcolor}" LINK="$conf{linkcolor}" ONLOAD="goToLogin()">
Sorry, but your session has either timed out, or has aborted due to error, returning you to the <A TARGET="_top" HREF="$Scriptname">Login page</A>
</BODY></HTML>
`;
        exit(0);
    }

    for my $a (keys %ENV) {
        $ENV{$a} =~ s/\n/ /g;
	print $s "$a: $ENV{$a}\n";
    }
    {
	local $/ = undef;
	print $s "\n";
	print $s <STDIN>;
    }
    while(my $bytecount = $s->read($_, $BufferSize)) { print; }
    $s->close();
    exit(0);
}
